"""Autogenerated DynamoDB exceptions.

This file was autogenerated by scripts/generate_exceptions.py.
Do not edit it manually!

"""
from dokklib_db_extended.errors.client import ClientError


class BackupInUseException(ClientError):
    """Please check DynamoDB docs for documentation."""


class BackupNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ConditionalCheckFailedException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ContinuousBackupsUnavailableException(ClientError):
    """Please check DynamoDB docs for documentation."""


class GlobalTableAlreadyExistsException(ClientError):
    """Please check DynamoDB docs for documentation."""


class GlobalTableNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class IdempotentParameterMismatchException(ClientError):
    """Please check DynamoDB docs for documentation."""


class IndexNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class InternalServerError(ClientError):
    """Please check DynamoDB docs for documentation."""


class InvalidRestoreTimeException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ItemCollectionSizeLimitExceededException(ClientError):
    """Please check DynamoDB docs for documentation."""


class LimitExceededException(ClientError):
    """Please check DynamoDB docs for documentation."""


class PointInTimeRecoveryUnavailableException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ProvisionedThroughputExceededException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ReplicaAlreadyExistsException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ReplicaNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class RequestLimitExceeded(ClientError):
    """Please check DynamoDB docs for documentation."""


class ResourceInUseException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ResourceNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TableAlreadyExistsException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TableInUseException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TableNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TransactionConflictException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TransactionInProgressException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ValidationError(ClientError):
    """Please check DynamoDB docs for documentation."""


class ThrottlingError(ClientError):
    """Please check DynamoDB docs for documentation."""



from abc import ABC, abstractmethod
from unittest.mock import MagicMock

from boto3.dynamodb.conditions import Key

from botocore.exceptions import ClientError

import dokklib_db_extended as db
from dokklib_db_extended.table import Table

from tests.unit import TestBase


class User(db.EntityName):
    pass


class Subscription(db.EntityName):
    pass


class TestRemoveEntityPrefix(TestBase):
    def _test_val(self, prefix, val):
        res = Table._remove_entity_prefix(f'{prefix}{val}')
        self.assertEqual(val, res)

    def test_noop_on_no_match(self):
        val = 'foo'
        res = Table._remove_entity_prefix(val)
        self.assertEqual(val, res)

    def test_removes_class_uppercase(self):
        prefix = 'A1B2_CD3#'
        val = 'foo'
        self._test_val(prefix, val)

    def test_handles_multiple_hashes(self):
        prefix = 'PREFIX#'
        val = '#foo#bar#'
        self._test_val(prefix, val)

    def test_handles_pipe(self):
        prefix = 'PREFIX#'
        val = 'foo|bar'
        self._test_val(prefix, val)


class TestStripPrefixes(TestBase):
    def setUp(self):
        self._pk = db.PartitionKey(User, 'foo@example.com')
        self._sk = db.SortKey(Subscription, 'docs.example.com')

    def test_noop_on_no_prefix(self):
        item = {
            'foo': 'bar'
        }
        res = Table._strip_prefixes(item)
        self.assertDictEqual(item, res)

    def test_strips_prefixes(self):
        item = {
            'PK': str(self._pk),
            'SK': str(self._sk),
            'Foo': str(self._sk)
        }
        res = Table._strip_prefixes(item)
        self.assertEqual(res['PK'], self._pk.value)
        self.assertEqual(res['SK'], self._sk.value)
        self.assertEqual(res['Foo'], self._sk.value)

    def test_makes_copy(self):
        item = {
            'PK': str(self._pk),
            'SK': str(self._sk)
        }
        res = Table._strip_prefixes(item)
        self.assertNotEqual(item['PK'], res['PK'])
        self.assertNotEqual(item['SK'], res['SK'])


class TestInit(TestBase):
    _to_patch = [
        'dokklib_db_extended.table.boto3'
    ]

    def test_client(self):
        boto3 = self._mocks['boto3']

        table = Table('my-table')
        self.assertEqual(table._client, boto3.client.return_value)

    def test_primary_index(self):
        pk_name = 'my-pk-name'
        sk_name = 'my-sk-name'

        class TestIndex(db.GlobalIndex):
            @property
            def partition_key(self) -> str:
                return pk_name

            @property
            def sort_key(self) -> str:
                return sk_name

        table = Table('my-table', TestIndex())
        self.assertEqual(table.primary_index.partition_key, pk_name)
        self.assertEqual(table.primary_index.sort_key, sk_name)


class TableTestCaseMixin(ABC):
    _to_patch = [
        'dokklib_db_extended.table.boto3',
        'dokklib_db_extended.table.Table._client#PROPERTY'
    ]

    @abstractmethod
    def _call_test_fn(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def _dynamo_method(self):
        raise NotImplementedError

    def setUp(self):
        super().setUp()

        self._client = MagicMock()
        self._mocks['_client'].return_value = self._client
        self._pk = db.PartitionKey(User, 'foo@example.com')
        self._sk = db.SortKey(Subscription, 'docs.example.com')
        self._sk_prefix = db.PrefixSortKey(Subscription)

    def test_handlers_throughput_error(self):
        error_response = {
            'Error': {
                'Code': 'ProvisionedThroughputExceededException',
            }
        }
        self._dynamo_method.side_effect = ClientError(error_response,
                                                      'OpName')
        with self.assertRaises(db.errors.ProvisionedThroughputExceededException):  # noqa 501
            self._call_test_fn()


class TestBatchGet(TableTestCaseMixin, TestBase):
    def setUp(self):
        super().setUp()
        self._pk_2 = db.PartitionKey(User, 'bar@example.com')
        self._sk_2 = db.SortKey(Subscription, 'docs.bar.com')
        self._keys = [
            db.PrimaryKey(self._pk, self._sk),
            db.PrimaryKey(self._pk_2, self._sk_2)
        ]
        self._table_name = 'my-table'
        self._table = Table(self._table_name)

    def _call_test_fn(self, attributes=None, consistent=False):
        attributes = attributes or []
        return self._table.batch_get(self._keys,
                                     attributes=attributes,
                                     consistent=consistent)

    def _get_call_arg(self, name, consistent=False, attributes=None):
        self._call_test_fn(attributes=attributes, consistent=consistent)
        _, kwargs = self._dynamo_method.call_args
        return kwargs['RequestItems'][self._table_name][name]

    def _get_attributes_call_arg(self, attributes=None):
        pe = self._get_call_arg('ProjectionExpression', attributes)
        attributes = pe.split(',')
        return attributes

    @property
    def _dynamo_method(self):
        return self._client.batch_get_item

    def test_keys(self):
        keys = self._get_call_arg('Keys', consistent=True)
        self.assertEqual(len(keys), 2)
        for i, key in enumerate(keys):
            exp = self._keys[i].serialize(self._table.primary_index)
            self.assertDictEqual(keys[i], exp)

    def test_consistent(self):
        consistent = self._get_call_arg('ConsistentRead', consistent=True)
        self.assertTrue(consistent)

    def test_retrieves_keys(self):
        attributes = self._get_attributes_call_arg(['foo'])
        self.assertTrue(set(attributes).issuperset({'PK', 'SK'}))

    def test_retrieves_keys_default(self):
        attributes = self._get_attributes_call_arg()
        self.assertSetEqual(set(attributes), {'PK', 'SK'})

    def test_results(self):
        key_1_ser = self._keys[0].serialize(self._table.primary_index)
        key_2_ser = self._keys[1].serialize(self._table.primary_index)
        self._dynamo_method.return_value = {
            'Responses': {
                self._table_name: [
                    key_1_ser
                ]
            },
            'UnprocessedKeys': {
                self._table_name: {
                    'Keys': [
                        key_2_ser
                    ]
                }
            }
        }
        res = self._call_test_fn()
        self.assertEqual(len(res.items), 1)
        self.assertEqual(res.items[0]['PK'], self._pk.value)
        self.assertEqual(res.items[0]['SK'], self._sk.value)

        self.assertEqual(len(res.unprocessed_keys), 1)
        self.assertEqual(res.unprocessed_keys[0], self._keys[1])


class TestDeleteItem(TableTestCaseMixin, TestBase):
    def _call_test_fn(self, table_name='my-table'):
        table = Table(table_name)
        return table.delete(self._pk, self._sk)

    @property
    def _dynamo_method(self):
        return self._client.delete_item


class QueryTestMixin(TableTestCaseMixin):
    def test_handles_no_result(self):
        self._dynamo_method.return_value = {}
        self.assertFalse(self._call_test_fn())

    def test_handles_empty_result(self):
        self._dynamo_method.return_value = {'Items': []}
        self.assertFalse(self._call_test_fn())


class TestQuery(QueryTestMixin, TestBase):
    def _call_test_fn(self, table_name='my-table'):
        table = Table(table_name)
        key_cond = Key('PK').eq(str(self._pk))
        query_arg = db.QueryArg(key_cond)
        return table._query(query_arg)

    @property
    def _dynamo_method(self):
        return self._client.query

    def test_strips_prefixes(self):
        self._dynamo_method.return_value = {
            'Items': [{'PK': {'S': str(self._pk)}}]
        }
        res = self._call_test_fn()
        self.assertEqual(res[0]['PK'], self._pk.value)


class TestGetItem(QueryTestMixin, TestBase):
    def _call_test_fn(self, attributes=None):
        table = Table('my-table')
        return table.get(self._pk, self._sk,
                         attributes=attributes)

    @property
    def _dynamo_method(self):
        return self._client.get_item

    def test_strips_prefixes(self):
        self._dynamo_method.return_value = {
            'Item': {'PK': {'S': str(self._pk)}}
        }
        res = self._call_test_fn()
        self.assertEqual(res['PK'], self._pk.value)


class TestQueryPrefix(QueryTestMixin, TestBase):
    def _call_test_fn(self, global_index=None, attributes=None):
        table = Table('my-table')
        return table.query_prefix(self._pk, self._sk_prefix,
                                  global_index=global_index,
                                  attributes=attributes)

    @property
    def _dynamo_method(self):
        return self._client.query

    def test_correct_key(self):
        self._call_test_fn()
        _, kwargs = self._dynamo_method.call_args
        kc = kwargs['KeyConditionExpression']
        self.assertEqual('(#n0 = :v0 AND begins_with(#n1, :v1))', kc)

    def test_global_index(self):
        index = db.InversePrimaryIndex()
        self._call_test_fn(global_index=index)
        _, kwargs = self._dynamo_method.call_args
        attr_names = kwargs['ExpressionAttributeNames']
        self.assertEqual(attr_names['#n0'], index.partition_key)
        self.assertEqual(attr_names['#n1'], index.sort_key)

    def test_defaults_to_global_index_sk_if_provided(self):
        index = db.InversePrimaryIndex()
        self._call_test_fn(global_index=index)
        _, kwargs = self._dynamo_method.call_args
        self.assertEqual(kwargs['ProjectionExpression'], index.sort_key)


class PutItemTestMixin(TableTestCaseMixin):
    def test_handles_conditional_check_failed(self):
        error_response = {'Error': {'Code': 'ConditionalCheckFailedException'}}
        self._dynamo_method.side_effect = ClientError(error_response,
                                                      'PutItem')
        with self.assertRaises(db.errors.ConditionalCheckFailedException):
            self._call_test_fn()


class TestPutItem(PutItemTestMixin, TestBase):
    def _call_test_fn(self, table_name='my-table'):
        table = Table(table_name)
        put_arg = db.PutArg(self._pk, self._sk)
        return table._put_item(put_arg)

    @property
    def _dynamo_method(self):
        return self._client.put_item


class TestTransactWriteItems(PutItemTestMixin, TestBase):

    def _call_test_fn(self, items=None, table_name='my-table'):
        table = Table(table_name)
        if not items:
            items = []
        return table.transact_write_items(items)

    @property
    def _dynamo_method(self):
        return self._client.transact_write_items

    def _setup_error(self, message=''):
        error_response = {
            'Error': {
                'Code': 'TransactionCanceledException',
                'Message': message
            }
        }
        self._dynamo_method.side_effect = ClientError(error_response,
                                                      'TransactWriteItems')

    def test_converts_to_op_name_dicts(self):
        op_name = 'my-op-name'
        table_name = 'foo-table-name'

        arg_mock = MagicMock(spec=db.PutArg)
        arg_mock.get_kwargs.return_value = 1
        arg_mock.op_name = op_name
        expected_item = {op_name: 1}

        self._call_test_fn(items=[arg_mock], table_name=table_name)
        arg_mock.get_kwargs.assert_called_once()
        args, _ = arg_mock.get_kwargs.call_args
        self.assertEqual(args[0], table_name)
        _, kwargs = self._dynamo_method.call_args
        self.assertDictEqual(kwargs, {'TransactItems': [expected_item]})

    def test_handles_transaction_failed(self):
        self._setup_error()
        with self.assertRaises(db.errors.TransactionCanceledException):
            self._call_test_fn()


class TestUpdateItem(TableTestCaseMixin, TestBase):
    def _call_test_fn(self, table_name='my-table'):
        table = Table(table_name)
        put_attributes = {
            'foo': 'bar'
        }
        update_arg = db.UpdateArg(self._pk, self._sk,
                                  attr_updates=put_attributes)
        return table._update_item(update_arg)

    @property
    def _dynamo_method(self):
        return self._client.update_item



from typing import Any, Dict


class ClientError(Exception):
    """Base class of Dokklib-DB client errors."""

    def __init__(self,
                 message: str,
                 error_response: Dict[str, Any],
                 operation_name: str):
        """Initialize a ClientError instance.

        Args:
            message: The error message.
            error_response: The error response dict from Boto.
            operation_name: The DynamoDB API operation name.

        """
        super().__init__(message)
        self.response = error_response
        self.operation_name = operation_name



"""DynamoDB operation arguments."""
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, List, Mapping, Optional, Union

import boto3.dynamodb.conditions as cond

from typing_extensions import Literal

from dokklib_db_extended.index import GlobalIndex, GlobalSecondaryIndex
from dokklib_db_extended.keys import PartitionKey, PrimaryKey, SortKey
from dokklib_db_extended.serializer import Serializer

_DynamoValue = Union[str, bool]

# Can't narrow value types down, because of TypedDict-Mapping
# incompatibiltiy. See https://stackoverflow.com/q/60304154
Attributes = Mapping[str, Any]
Kwargs = Mapping[str, Any]


class OpArg(ABC):
    """DynamoDB operation argument base class."""

    @staticmethod
    def _iso_now() -> str:
        now = datetime.utcnow()
        return now.replace(microsecond=0).isoformat()

    def __init__(self) -> None:
        """Initialize an OpArg instance."""
        self._serializer = Serializer()

    @property
    @abstractmethod
    def op_name(self) -> str:
        """Get the operation name for which this object is an argument.

        Must correspond to TransactWriteItem argument.
        """
        raise NotImplementedError

    @abstractmethod
    def get_kwargs(self, table_name: str, primary_index: GlobalIndex) \
            -> Kwargs:
        """Get key-word arguments that can be passed to the DynamoDB operation.

        Args:
            table_name: The DynamoDB table name for the operation.
            primary_index: The primary global index of the table.

        Returns:
            The key-word arguments.

        """
        raise NotImplementedError

    def _serialize_primary_key(self, primary_index: GlobalIndex,
                               pk: PartitionKey, sk: SortKey) \
            -> Mapping[str, Mapping[str, _DynamoValue]]:
        """Serialize composite key."""
        primary_key = PrimaryKey(pk, sk)
        return primary_key.serialize(primary_index)


class DeleteArg(OpArg):
    """Argument to a DynamoDB DeleteItem operation."""

    def __init__(self, pk: PartitionKey, sk: SortKey,
                 idempotent: bool = True):
        """Initialize a DeleteArg instance.

        Args:
            pk: The partition key of the item.
            sk: The sort key of the item.
            idempotent: If false, the op raises an error if the item to
                delete doesn't exist. Defaults to to true.

        """
        super().__init__()
        self._pk = pk
        self._sk = sk
        self._idempotent = idempotent

    @property
    def op_name(self) -> Literal['Delete']:  # pragma: no cover
        """Get the operation name for which this object is an argument."""
        return 'Delete'

    def get_kwargs(self, table_name: str, primary_index: GlobalIndex) \
            -> Kwargs:
        """Get key-word arguments that can be passed to a DeleteItem operation.

        Args:
            table_name: The DynamoDB table name for the DeleteItem operation.
            primary_index: The primary global index of the table.

        Returns:
            The key-word arguments.

        """
        key = self._serialize_primary_key(primary_index, self._pk, self._sk)
        kwargs = {
            'TableName': table_name,
            'Key': key
        }
        if not self._idempotent:
            # This check is performed after the item is retrieved by the
            # composite key, so no need to specify SK.
            kwargs['ConditionExpression'] = 'attribute_exists(PK)'

        return kwargs


class GetArg(OpArg):
    """Argument to a DynamoDB GetItem operation."""

    def __init__(self, pk: PartitionKey, sk: SortKey,
                 attributes: Optional[List[str]] = None,
                 consistent: bool = False):
        """Initialize a GetArg instance.

        Args:
            pk: The partition key of the item.
            sk: The sort key of the item.
            attributes: The attributes to get. Returns all attributes if
                omitted.
            consistent: Whether the read is strongly consistent or not.

        """
        super().__init__()
        self._pk = pk
        self._sk = sk
        self._attributes = attributes
        self._consistent = consistent

    @property
    def op_name(self) -> Literal['Get']:  # pragma: no cover
        """Get the operation name for which this object is an argument."""
        return 'Get'

    def get_kwargs(self, table_name: str, primary_index: GlobalIndex) \
            -> Kwargs:
        """Get key-word arguments that can be passed to a GetItem operation.

        Args:
            table_name: The DynamoDB table name for the GetItem operation.
            primary_index: The primary global index of the table.

        Returns:
            The key-word arguments.

        """
        key = self._serialize_primary_key(primary_index, self._pk, self._sk)
        kwargs = {
            'TableName': table_name,
            'Key': key,
            'ConsistentRead': self._consistent
        }
        if self._attributes:
            # TODO (abiro) convert inputs to expression attribute names
            kwargs['ProjectionExpression'] = ','.join(self._attributes)

        return kwargs


class PutArg(OpArg):
    """Argument to a DynamoDB PutItem operation.

    This op will replace the entire item. User `UpdateArg` if you just want to
    update a specific attribute.

    The `CreatedAt` attribute of the item is automatically set to the current
    ISO timestamp without microseconds (eg. '2020-02-15T19:09:38').

    """

    def __init__(self, pk: PartitionKey, sk: SortKey,
                 attributes: Optional[Attributes] = None,
                 allow_overwrite: bool = True):
        """Initialize a PutArg instance.

        Args:
            pk: The partition key of the item.
            sk: The sort key of the item.
            attributes: Optional additional attributes of the item.
            allow_overwrite: Whether to allow overwriting an existing item.

        """
        super().__init__()
        self._pk = pk
        self._sk = sk
        self._attributes = attributes
        self._allow_overwrite = allow_overwrite

    @property
    def op_name(self) -> Literal['Put']:  # pragma: no cover
        """Get the operation name for which this object is an argument."""
        return 'Put'

    def _get_dynamo_item(self, primary_index: GlobalIndex) \
            -> Mapping[str, Mapping[str, _DynamoValue]]:
        keys_item = self._serialize_primary_key(primary_index,
                                                self._pk,
                                                self._sk)

        item: Attributes = {
            'CreatedAt': self._iso_now()
        }
        if self._attributes:
            # `item` keys overwrite `_attributes` keys
            item = {**self._attributes, **item}

        dynamo_item = self._serializer.serialize_dict(item)
        return {**dynamo_item, **keys_item}

    def get_kwargs(self, table_name: str, primary_index: GlobalIndex) \
            -> Kwargs:
        """Get key-word arguments that can be passed to a PutItem operation.

        Args:
            table_name: The DynamoDB table name for the PutItem operation.
            primary_index: The primary global index of the table.

        Returns:
            The key-word arguments.

        """
        kwargs = {
            'TableName': table_name,
            'Item': self._get_dynamo_item(primary_index)
        }
        if not self._allow_overwrite:
            # The condition only checks if the item with the same composite key
            # exists. Ie. if there is an item (PK=foo, SK=0) in the table,
            # and we insert a new item (PK=foo, SK=1), the insert will succeed.
            kwargs['ConditionExpression'] = 'attribute_not_exists(PK)'

        return kwargs


class InsertArg(PutArg):
    """DynamoDB PutItem argument that prevents overwriting existing items."""

    def __init__(self, pk: PartitionKey, sk: SortKey,
                 attributes: Optional[Attributes] = None):
        """Initialize an InsertArg instance.

        The `UpdateAt` attribute of the item is automatically set.

        Args:
            pk: The partition key of the item.
            sk: The sort key of the item.
            attributes: Optional additional attributes of the item.

        """
        super().__init__(pk, sk,
                         attributes=attributes,
                         allow_overwrite=False)


class QueryArg(OpArg):
    """DynamoDB query operation argument.

    Note that query can not be used in a transaction, the purpose of this class
    is to provide a simplified interface to the boto3 DynamoDB table class.

    """

    _max_limit = 1000

    @staticmethod
    def _serialize_key_condition(key_cond: cond.ConditionBase) \
            -> cond.BuiltConditionExpression:
        builder = cond.ConditionExpressionBuilder()
        return builder.build_expression(key_cond, is_key_condition=True)

    def __init__(self, key_condition: cond.ConditionBase,
                 global_index: Optional[GlobalSecondaryIndex] = None,
                 attributes: Optional[List[str]] = None,
                 consistent: bool = False,
                 limit: Optional[int] = None):
        """Initialize a QueryArg instance.

        Args:
            key_condition: The key condition. Eg.:
                `Key('PK').eq(str(pk)) & Key('SK').begins_with(str(sk))`
            global_index: The global secondary index to query. Defaults to the
                primary index.
            attributes: The attributes to get. Defaults to `SK`.
            consistent: Whether the read is strongly consistent or not.
            limit: The maximum number of items to fetch. Defaults to 1000 which
                is also the maximum allowed value.

        """
        super().__init__()
        self._key_cond = key_condition
        self._attributes = attributes
        self._consistent = consistent
        self._global_index = global_index

        if limit is not None:
            if limit > self._max_limit:
                raise ValueError(f'Limit {limit} is greater than max '
                                 f'{self._max_limit}')
            self._limit = limit
        else:
            self._limit = self._max_limit

    @property
    def op_name(self) -> Literal['Query']:  # pragma: no cover
        """Get the operation name for which this object is an argument.

        Note that query can not be used in a transaction.

        """
        return 'Query'

    def _serialize_primary_key(self, primary_index: GlobalIndex,
                               pk: PartitionKey,
                               sk: SortKey) \
            -> Mapping[str, Mapping[str, _DynamoValue]]:
        """Serialize composite key."""
        # Using this inherited method in QueryArg would be a mistake, because
        # it wouldn't take into account the global secondary index.
        raise NotImplementedError

    def get_kwargs(self, table_name: str, primary_index: GlobalIndex) \
            -> Kwargs:
        """Get key-word arguments that can be passed to a boto3 DynamoDB table.

        Args:
            table_name: The DynamoDB table name for the operation.
            primary_index: The primary global index of the table.

        Returns:
            The key-word arguments.

        """
        kc = self._serialize_key_condition(self._key_cond)
        kc_value_placeholders = self._serializer.serialize_dict(
            kc.attribute_value_placeholders)
        kwargs = {
            'TableName': table_name,
            'Select': 'SPECIFIC_ATTRIBUTES',
            'KeyConditionExpression': kc.condition_expression,
            'ExpressionAttributeNames': kc.attribute_name_placeholders,
            'ExpressionAttributeValues': kc_value_placeholders,
            'ConsistentRead': self._consistent,
            'Limit': self._limit
        }
        if self._attributes:
            # TODO (abiro) convert inputs to expression attribute names
            kwargs['ProjectionExpression'] = ','.join(self._attributes)
        else:
            kwargs['ProjectionExpression'] = 'SK'
        if self._global_index:
            kwargs['IndexName'] = self._global_index.name
        return kwargs


class UpdateArg(OpArg):
    """Argument to a DynamoDB UpdateItem operation.

    This op updates the specified attributes or creates a new item if it
    doesn't exist yet.

    The `UpdatedAt` attribute of the item is automatically set to the current
    ISO timestamp without microseconds (eg. '2020-02-15T19:09:38').

    """

    def __init__(self, pk: PartitionKey, sk: SortKey,
                 attr_updates: Optional[Attributes] = None):
        """Initialize an UpdateArg instance.

        Args:
            pk: The partition key of the item.
            sk: The sort key of the item.
            attr_updates: Optional attributes to update for the item. These
                attributes will be overwritten if they exist, or created if
                they don't exist.

        """
        super().__init__()
        self._pk = pk
        self._sk = sk
        self._attr_updates = attr_updates

    @property
    def op_name(self) -> Literal['Update']:  # pragma: no cover
        """Get the operation name for which this object is an argument."""
        return 'Update'

    def _get_attr_updates(self) -> Mapping[str, Mapping[str, Any]]:
        item = {
            'UpdatedAt': self._iso_now()
        }
        if self._attr_updates:
            # `item` keys overwrite `_attributes` keys
            item = {**self._attr_updates, **item}
        res = {}
        for k, v in item.items():
            res[k] = {
                'Action': 'PUT',
                'Value': self._serializer.serialize_val(v)
            }
        return res

    def get_kwargs(self, table_name: str, primary_index: GlobalIndex) \
            -> Kwargs:
        """Get key-word arguments that can be passed to a PutItem operation.

        Args:
            table_name: The DynamoDB table name for the PutItem operation.
            primary_index: The primary global index of the table.

        Returns:
            The key-word arguments.

        """
        keys = self._serialize_primary_key(primary_index, self._pk, self._sk)
        attr_updates = self._get_attr_updates()
        kwargs = {
            'TableName': table_name,
            'Key': keys,
            'AttributeUpdates': attr_updates
        }
        return kwargs



from typing import Any, Dict, Mapping, Optional, cast

from boto3.dynamodb.types import TypeDeserializer, TypeSerializer


_StrKeyDict = Dict[str, Any]


class Serializer:
    """Convert between Python and DynamoDB values."""

    def __init__(self) -> None:
        """Initialize a Serializer instance."""
        # Lazy-initialize serializers
        self._ser_handle: Optional[TypeSerializer] = None
        self._deser_handle: Optional[TypeDeserializer] = None

    @property
    def _deser(self) -> TypeDeserializer:
        if self._deser_handle is None:
            self._deser_handle = TypeDeserializer()
        return self._deser_handle

    @property
    def _ser(self) -> TypeSerializer:
        if self._ser_handle is None:
            self._ser_handle = TypeSerializer()
        return self._ser_handle

    def deserialize_dict(self, dynamo_item: Mapping[str, Any]) -> _StrKeyDict:
        """Deserialize a dictionary while preserving its top level keys.

        Args:
            dynamo_item: The dictionary from DynamoDB to deserialize.

        Returns:
            The deserialized dictionary.

        Raises:
            TypeError: if an unsupported type is encountered.

        """
        return {k: self.deserialize_val(v) for k, v in dynamo_item.items()}

    def deserialize_val(self, dynamo_val: Mapping[str, Any]) -> Any:
        """Convert a DynamoDB value to a Python value.

        Args:
            dynamo_val: The value from DynamoDB to deserialize.

        Returns:
            The deserialized value.

        Raises:
            TypeError: if an unsupported type is encountered.

        """
        return cast(_StrKeyDict, self._deser.deserialize(dynamo_val))

    def serialize_val(self, val: Any) -> _StrKeyDict:
        """Convert a Python value to a DynamoDB value.

        Args:
            val: The value to serialize.

        Returns:
            The serialized DynamoDB value.

        Raises:
            TypeError: if an unsupported type is encountered.

        """
        return cast(_StrKeyDict, self._ser.serialize(val))

    def serialize_dict(self, item: Mapping[str, Any]) -> _StrKeyDict:
        """Serialize a dictionary while preserving its top level keys.

        Args:
            item: The dictionary to serialize.

        Returns:
            The serialized DynamoDB dictionary.

        Raises:
            TypeError: if an unsupported type is encountered.

        """
        return {k: self.serialize_val(v) for k, v in item.items()}



# flake8: noqa
# mypy: implicit-reexport

# Flake 8 would complain about unused imports if it was enabled on this file.

import dokklib_db_extended.errors

from dokklib_db_extended.table import (
    Table,
    ItemResult
)
from dokklib_db_extended.index import (
    GlobalIndex,
    GlobalSecondaryIndex,
    InversePrimaryIndex,
    PrimaryGlobalIndex
)
from dokklib_db_extended.keys import (
    AnySortKey,
    EntityName,
    PartitionKey,
    PrefixSortKey,
    PrimaryKey,
    SortKey,
)
from dokklib_db_extended.op_args import (
    Attributes,
    DeleteArg,
    GetArg,
    InsertArg,
    OpArg,
    PutArg,
    QueryArg,
    UpdateArg
)



# Type checks are enough to test this module.
# pragma: no cover
from abc import ABC, abstractmethod

from typing_extensions import Literal


class GlobalIndex(ABC):
    """Base class for global indices."""

    @property
    @abstractmethod
    def partition_key(self) -> str:
        """Get the name of the partition key."""
        raise NotImplementedError

    @property
    @abstractmethod
    def sort_key(self) -> str:
        """Get the name of the sort key."""
        raise NotImplementedError


class GlobalSecondaryIndex(GlobalIndex):
    """Base class for global secondary indices."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Get the name of the secondary index."""
        raise NotImplementedError


class PrimaryGlobalIndex(GlobalIndex):
    """Primary global DynamoDB primary_index."""

    @property
    def partition_key(self) -> Literal['PK']:
        """Get the name of the partition key."""
        return 'PK'

    @property
    def sort_key(self) -> Literal['SK']:
        """Get the name of the sort key."""
        return 'SK'


class InversePrimaryIndex(GlobalSecondaryIndex):
    """Inverted global secondary index."""

    @property
    def name(self) -> Literal['GSI_1']:
        """Get the global secondary index name."""
        return 'GSI_1'

    @property
    def partition_key(self) -> Literal['SK']:
        """Get the name of the partition key."""
        return 'SK'

    @property
    def sort_key(self) -> Literal['PK']:
        """Get the name of the sort key."""
        return 'PK'



import copy
import re
from contextlib import contextmanager
from typing import Any, Dict, Iterable, Iterator, List, Mapping, NamedTuple, \
    Optional, Tuple, Type, Union, cast

import boto3
import boto3.dynamodb.conditions as cond

import botocore.client
import botocore.exceptions as botoex

import dokklib_db_extended.errors as err
from dokklib_db_extended.index import GlobalIndex, GlobalSecondaryIndex, \
    PrimaryGlobalIndex
from dokklib_db_extended.keys import PartitionKey, PrefixSortKey, PrimaryKey, SortKey
from dokklib_db_extended.op_args import Attributes, DeleteArg, GetArg, InsertArg, \
    OpArg, PutArg, QueryArg, UpdateArg
from dokklib_db_extended.serializer import Serializer

ItemResult = Mapping[str, Any]


class BatchGetResult(NamedTuple):
    """Result from a `Table.batch_get` operation."""

    items: List[ItemResult]
    unprocessed_keys: List[PrimaryKey]


class Table:
    """DynamoDB table for the single table pattern.

    Table instances are not safe to share across threads.
    """

    @staticmethod
    def _get_error_code(error: botoex.ClientError) -> str:
        db_error = error.response.get('Error', {})
        return cast(str, db_error.get('Code', 'None'))

    @staticmethod
    def _remove_entity_prefix(string: str) -> str:
        # Entity names are upper-cased Python class names.
        pattern = r'^[A-Z0-9_]+#(.+)$'
        match = re.match(pattern, string)
        if match:
            return match.group(1)
        else:
            return string

    @classmethod
    @contextmanager
    def _dispatch_transaction_error(cls, op_args: List[OpArg]) \
            -> Iterator[None]:
        """Raise appropriate exception based on ClientError code."""
        try:
            yield None
        except botoex.ClientError as e:
            code = cls._get_error_code(e)
            if code == 'TransactionCanceledException':
                raise err.TransactionCanceledException(op_args,
                                                       str(e),
                                                       e.response,
                                                       e.operation_name)
            else:
                raise cls._get_exception(e)

    @classmethod
    @contextmanager
    def _dispatch_error(cls) -> Iterator[None]:
        """Raise appropriate exception based on ClientError code."""
        try:
            yield None
        except botoex.ClientError as e:
            raise cls._get_exception(e)

    @classmethod
    def _get_exception(cls, error: botoex.ClientError) -> err.ClientError:
        code = cls._get_error_code(error)
        try:
            ex_class = cast(Type[err.ClientError], getattr(err, code))
        except AttributeError:  # pragma: no cover
            # Type checks are enough to test this.
            ex_class = err.ClientError
        return ex_class(str(error), error.response, error.operation_name)

    @classmethod
    def _strip_prefixes(cls, item: Dict[str, Any]) -> ItemResult:
        """Strip entity prefixes from a DB item."""
        item_copy = copy.deepcopy(item)
        for k, v in item_copy.items():
            if isinstance(v, str):
                item_copy[k] = cls._remove_entity_prefix(v)
        return item_copy

    def __init__(self, table_name: str,
                 primary_index: Optional[GlobalIndex] = None,
                 region_name: Optional[str] = None,
                 endpoint_url: Optional[str] = None,
                 use_ssl: Optional[str] = None,
                 aws_access_key_id: Optional[str] = None,
                 aws_secret_access_key: Optional[str] = None):
        """Initialize a Table instance.

        Args:
            table_name: The DynamoDB table name.
            primary_index: The primary global index of the table.
                Defaults to `db.PrimaryGlobalIndex` that has 'PK' as the
                partition key name and 'SK' as the sort key name.

        """
        self._table_name = table_name
        if primary_index:
            self._primary_index = primary_index
        else:
            self._primary_index = PrimaryGlobalIndex()
        self._serializer = Serializer()

        # The boto objects are lazy-initialzied. Connections are not created
        # until the first request.
        session = boto3.session.Session()
        credentials = session.get_credentials()
        self._client_handle = boto3.client('dynamodb',
                                           endpoint_url=endpoint_url,
                                           region_name=region_name or session.region_name,
                                           use_ssl=use_ssl or True,
                                           aws_access_key_id=aws_access_key_id or credentials.access_key,
                                           aws_secret_access_key=aws_secret_access_key or credentials.secret_key,
                                           )

    @property
    def _client(self) -> 'botocore.client.DynamoDB':
        # Helps mock the client at test time.
        return self._client_handle

    @property
    def primary_index(self) -> GlobalIndex:
        """Get the primary global index of the table."""
        return self._primary_index

    @property
    def table_name(self) -> str:
        """Get the DynamoDB table name."""
        return self._table_name

    def _normalize_item(self, item: Dict[str, Any]) -> ItemResult:
        des_item = self._serializer.deserialize_dict(item)
        return self._strip_prefixes(des_item)

    def _normalize_items(self, items: List[Dict[str, Any]]) \
            -> List[ItemResult]:
        return [self._normalize_item(item) for item in items]

    def _put_item(self, put_arg: PutArg) -> None:
        kwargs = put_arg.get_kwargs(self.table_name, self.primary_index)
        with self._dispatch_error():
            self._client.put_item(**kwargs)

    def _query(self, query_arg: QueryArg) -> List[ItemResult]:
        args = query_arg.get_kwargs(self.table_name, self.primary_index)
        with self._dispatch_error():
            query_res = self._client.query(**args)
        items = query_res.get('Items', [])
        return self._normalize_items(items)

    def _update_item(self, update_arg: UpdateArg) -> None:
        """Update an item or insert a new item if it doesn't exist.

        Args:
            update_arg: The update item op argument.

        Raises:
            dokklib_db_extended.DatabaseError if there was a problem connecting to
                DynamoDB.

        """
        kwargs = update_arg.get_kwargs(self.table_name, self.primary_index)
        with self._dispatch_error():
            self._client.update_item(**kwargs)

    def batch_get(self, keys: Iterable[PrimaryKey],
                  attributes: Optional[List[str]] = None,
                  consistent: bool = False) -> BatchGetResult:
        """Fetch multiple items by their primary keys from the table.

        Note that the Dynamodb BatchGetItem API operation doesn't return items
        in order, that's why the primary key (PK and SK) of the item is always
        included in the Table.batch_get results.

        Further, note that while it's possible to make indiviual reads in
        strongly consistent, the returned snapshot has no isolation guarantees.
        If you need a consistent snapshot of multiple items in the database,
        you should use a transaction.

        Doesn't handle `UnprocessedKeys` in response.

        Args:
            keys: The primary keys of the items to get.
            attributes: The attributes to get. Returns all attributes if
                omitted. The partition and sort keys are always included even
                if not specified here.
            consistent: Whether the read is strongly consistent or not.

        Returns:
            The item if it exists.

        """
        attr_s = set(attributes or [])
        attr_s.add(self.primary_index.partition_key)
        attr_s.add(self.primary_index.sort_key)
        # TODO (abiro) convert inputs to expression attribute names
        proj_expr = ','.join(attr_s)

        key_map: Dict[Union[PrimaryKey, Tuple[str, str]], PrimaryKey] = {}
        key_items = []
        for key in keys:
            key_map[key] = key
            ser_key = key.serialize(self.primary_index)
            key_items.append(ser_key)

        request_items = {
            self.table_name: {
                'Keys': key_items,
                'ProjectionExpression': proj_expr,
                'ConsistentRead': consistent
            }
        }
        with self._dispatch_error():
            res = self._client.batch_get_item(RequestItems=request_items)

        responses = res.get('Responses', {})
        items = responses.get(self.table_name, [])
        norm_items = self._normalize_items(items)

        # Map unprocessed keys back to original `PrimaryKey` arguments.
        unproc = res.get('UnprocessedKeys', {})
        unproc_items = unproc.get(self.table_name, {})
        unproc_keys = []
        for item in unproc_items.get('Keys', []):
            pk_dynamo = item[self.primary_index.partition_key]
            sk_dynamo = item[self.primary_index.sort_key]
            pk_val = self._serializer.deserialize_val(pk_dynamo)
            sk_val = self._serializer.deserialize_val(sk_dynamo)
            key_tuple = (cast(str, pk_val), cast(str, sk_val))
            key = key_map[key_tuple]
            unproc_keys.append(key)

        return BatchGetResult(items=norm_items, unprocessed_keys=unproc_keys)

    def delete(self, pk: PartitionKey, sk: SortKey,
               idempotent: bool = True) -> None:
        """Delete an item from the table.

        Args:
            pk: The primary key.
            sk: The sort key.
            idempotent: Whether the operation is idempotent. Defaults to True.

        """
        delete_arg = DeleteArg(pk, sk, idempotent=idempotent)
        kwargs = delete_arg.get_kwargs(self.table_name, self.primary_index)
        with self._dispatch_error():
            self._client.delete_item(**kwargs)

    def get(self, pk: PartitionKey, sk: SortKey,
            attributes: Optional[List[str]] = None,
            consistent: bool = False) -> Optional[ItemResult]:
        """Fetch an item by its primary key from the table.

        Args:
            pk: The primary key.
            sk: The sort key.
            attributes: The attributes to get. Returns all attributes if
                omitted.
            consistent: Whether the read is strongly consistent or not.

        Returns:
            The item if it exists.

        """
        get_arg = GetArg(pk, sk, attributes=attributes, consistent=consistent)
        kwargs = get_arg.get_kwargs(self.table_name, self.primary_index)
        with self._dispatch_error():
            res = self._client.get_item(**kwargs)
        item = res.get('Item')
        if item:
            return self._normalize_item(item)
        else:
            return None

    # Type checks are sufficient to test this function, so it's excluded from
    # unit test coverage.
    def insert(self, pk: PartitionKey, sk: SortKey,
               attributes: Optional[Attributes] = None) -> None:  # pragma: no cover  # noqa 501
        """Insert a new item into the table.

        The UpdateAt attribute of the item is automatically set.
        The insert fails if an item with the same composite key (PK, SK)
        exists.

        Args:
            pk: The partition key.
            sk: The sort key.
            attributes: Dictionary with additional attributes of the item.

        Raises:
            dokklib_db_extended.ItemExistsError if the item with the same composite
                key already exists.
            dokklib_db_extended.DatabaseError if there was a problem connecting to
                DynamoDB.

        """
        put_arg = InsertArg(pk, sk, attributes=attributes)
        self._put_item(put_arg)

    # Type checks are sufficient to test this function, so it's excluded from
    # unit test coverage.
    def put(self, pk: PartitionKey, sk: SortKey,
            attributes: Optional[Attributes] = None,
            allow_overwrite: bool = True) -> None:  # pragma: no cover  # noqa 501
        """Insert a new item or replace an existing item.

        Args:
            pk: The partition key of the item.
            sk: The sort key of the item.
            attributes: Optional additional attributes of the item.
            allow_overwrite: Whether to allow overwriting an existing item.

        Raises:
            dokklib_db_extended.DatabaseError if there was a problem connecting to
                DynamoDB.

        """
        put_arg = PutArg(pk, sk,
                         attributes=attributes,
                         allow_overwrite=allow_overwrite)
        self._put_item(put_arg)

    # Type checks are sufficient to test this function, so it's excluded from
    # unit test coverage.
    def query(self, key_condition: cond.ConditionBase,
              global_index: Optional[GlobalSecondaryIndex] = None,
              attributes: Optional[List[str]] = None,
              consistent: bool = False,
              limit: Optional[int] = None) -> List[ItemResult]:  # pragma: no cover  # noqa 501
        """Fetch items from the table based on a key condition.

        Doesn't support pagination.

        Args:
            key_condition: The key condition. Eg.:
                `Key('PK').eq(str(pk)) & Key('SK').begins_with(str(sk))`
            global_index: The global secondary index to query. Defaults to the
                primary index.
            attributes: The attributes to get. Defaults to `SK`.
            consistent: Whether the read is strongly consistent or not.
            limit: The maximum number of items to fetch. Defaults to 1000.

        Returns:
            The requested items with the entity name prefixes stripped,
            eg. if the value of an attribute is 'USER#foo@example.com',
            only 'foo@example.com' is returned.

        Raises:
            dokklib_db_extended.DatabaseError if there was an error querying the
                table.

        """
        query_arg = QueryArg(key_condition,
                             global_index=global_index,
                             attributes=attributes,
                             consistent=consistent,
                             limit=limit)
        return self._query(query_arg)

    def query_prefix(self, pk: PartitionKey, sk: PrefixSortKey,
                     global_index: Optional[GlobalSecondaryIndex] = None,
                     attributes: Optional[List[str]] = None,
                     consistent: bool = False,
                     limit: Optional[int] = None) -> List[ItemResult]:
        """Fetch a items from the table based on a sort key prefix.

        Doesn't support pagination.

        Args:
            pk: The partition key.
            sk: The sort key prefix.
            global_index: The global secondary index to query. Defaults to the
                primary index.
            attributes: The attributes to get. Defaults to
                `[self.primary_index.sort_key]` if no `global_index` is
                provided and `[global_index.sort_key]` if it is provided.
            consistent: Whether the read is strongly consistent or not.
            limit: The maximum number of items to fetch. Defaults to 1000.

        Returns:
            The requested items with the `PK` and `SK` prefixes stripped.

        Raises:
            dokklib_db_extended.DatabaseError if there was an error querying DynamoDB.

        """
        if global_index:
            pk_name = global_index.partition_key
            sk_name = global_index.sort_key
        else:
            pk_name = self.primary_index.partition_key
            sk_name = self.primary_index.sort_key

        if not attributes:
            attributes = [sk_name]

        key_condition = cond.Key(pk_name).eq(str(pk)) & \
                        cond.Key(sk_name).begins_with(str(sk))
        query_arg = QueryArg(key_condition,
                             global_index=global_index,
                             attributes=attributes,
                             consistent=consistent,
                             limit=limit)
        return self._query(query_arg)

    def transact_write_items(self, op_args: List[OpArg]) -> None:
        """Write multiple items in a transaction.

        Args:
            op_args: Write operation arguments.

        Raises:
            dokklib_db_extended.TransactionError if the transaction fails.
            dokklib_db_extended.DatabaseError if there was a problem connecting
                DynamoDB.

        """
        transact_items = []
        for a in op_args:
            kwargs = a.get_kwargs(self.table_name, self.primary_index)
            transact_items.append({a.op_name: kwargs})
        with self._dispatch_transaction_error(op_args):
            self._client.transact_write_items(TransactItems=transact_items)

    # Type checks are sufficient to test this function, so it's excluded from
    # unit test coverage.
    def update_attributes(self, pk: PartitionKey, sk: SortKey,
                          attributes: Attributes) -> None:  # pragma: no cover
        """Update an item or insert a new item if it doesn't exist.

        The `UpdatedAt` attribute of the item is automatically set.

        Args:
            pk: The partition key.
            sk: The sort key.
            attributes: Dictionary with attributes to updates. These attributes
                will overwritten if they exist or created if they don't exist.

        Raises:
            dokklib_db_extended.DatabaseError if there was a problem connecting to
                DynamoDB.

        """
        update_arg = UpdateArg(pk, sk, attr_updates=attributes)
        self._update_item(update_arg)



#!/usr/bin/env python3
"""DynamoDB integration tests.

!!! This script will clear the table first, so make sure to only use it on a
testing table. !!!
"""
import logging

import boto3

import dokklib_db_extended as db


TABLE_NAME = 'DokklibDB-IntegrationTest-SingleTable'

logging.basicConfig(level=logging.INFO)


class Order(db.EntityName):
    """Order entity name.

    Key value: unique order id with date prefixed, eg: '2020-02-21|order-id'
    Example key: 'ORDER#2020-02-21|order-id'.

    """


class User(db.EntityName):
    """User entity name.

    Key value: user email, eg: 'alice@example.com'.
    Example key: 'USER#alice@example.com'.

    """


class Product(db.EntityName):
    """Order entity name.

    Key value: unique product name, eg: 'my-book'.
    Example key: 'PRODUCT#my-book'.

    """


# From: https://stackoverflow.com/a/56616499
def _clear_db(table_name: str):
    logging.info('Clearing table')
    table = boto3.resource('dynamodb').Table(table_name)
    scan = None

    with table.batch_writer() as batch:
        while scan is None or 'LastEvaluatedKey' in scan:
            if scan is not None and 'LastEvaluatedKey' in scan:
                scan = table.scan(
                    ProjectionExpression='PK,SK',
                    ExclusiveStartKey=scan['LastEvaluatedKey'],
                )
            else:
                scan = table.scan(ProjectionExpression='PK,SK')

            for item in scan['Items']:
                batch.delete_item(Key={'PK': item['PK'], 'SK': item['SK']})


logging.info('Starting integration tests')

# We clear the DB instead of recreating it to save time.
_clear_db(TABLE_NAME)
table = db.Table(TABLE_NAME)

# Users
pk_alice = db.PartitionKey(User, 'alice@example.com')
sk_alice = db.SortKey(User, 'alice@example.com')

# Products
pk_book = db.PartitionKey(Product, 'book')

# Orders
pk_order1 = db.PartitionKey(Order, '2020-02-21|order-1')
sk_order1 = db.SortKey(Order, '2020-02-21|order-1')
sk_order2 = db.SortKey(Order, '2020-02-21|order-2')

logging.info('Testing insert')
table.insert(pk_alice, sk_alice)

logging.info('Testing update_attributes')
table.update_attributes(pk_alice, sk_alice, {'MyJson': {'A': 1}})

logging.info('Testing get_item')
res = table.get(pk_alice, sk_alice,
                attributes=['MyJson'],
                consistent=True)
assert res['MyJson']['A'] == 1, res

logging.info('Testing transact_write_items')
table.transact_write_items([
    db.InsertArg(pk_alice, sk_order1),
    db.InsertArg(pk_book, sk_order1),
    db.InsertArg(pk_alice, sk_order2)
])

logging.info('Testing transact_write_items error handling')
try:
    table.transact_write_items([
        db.InsertArg(pk_alice, sk_order1),
        db.InsertArg(pk_book, sk_order2)
    ])
except db.errors.TransactionCanceledException as e:
    assert e.reasons[0] is db.errors.ConditionalCheckFailedException, e.reasons
    assert e.reasons[1] is None, e.reasons

logging.info('Testing batch_get')
res = table.batch_get([
    db.PrimaryKey(pk_alice, sk_order1),
    db.PrimaryKey(pk_book, sk_order1),
    db.PrimaryKey(pk_alice, sk_order2),
], consistent=True)
assert len(res.items) == 3, res

logging.info('Testing query_prefix')
res = table.query_prefix(pk_alice, db.PrefixSortKey(Order))
assert len(res) == 2, res

# logging.info('Testing query_prefix on inverse index')
# res = table.query_prefix(pk_order1, db.PrefixSortKey(User),
#                          global_index=db.InversePrimaryIndex())
# assert len(res) == 1, res

logging.info('Testing delete')
table.delete(pk_alice, sk_alice, idempotent=False)

logging.info('+++ Success +++')



import dokklib_db_extended.serializer as m

from tests.unit import TestBase


class TestSerialize(TestBase):
    def setUp(self):
        self._ser = m.Serializer()

    def test_deserialize_dict(self):
        d = {
            'foo': {'S': 'bar'},
            'baz': {'N': '1'}
        }
        exp = {
            'foo': 'bar',
            'baz': 1
        }
        res = self._ser.deserialize_dict(d)
        self.assertDictEqual(exp, res)

    def test_deserialize_val(self):
        val = {'L': [{'N': '1'}, {'N': '2'}]}
        res = self._ser.deserialize_val(val)
        self.assertListEqual(res, [1, 2])

    def test_serialize_dict(self):
        d = {
            'foo': False,
            'bar': 1
        }
        res = self._ser.serialize_dict(d)
        exp = {
            'foo': {'BOOL': False},
            'bar': {'N': '1'}
        }
        self.assertDictEqual(res, exp)

    def test_serialize_val(self):
        s = {'1', '2'}
        res = self._ser.serialize_val(s)
        self.assertSetEqual(set(res['SS']), s)



from typing import Dict, List
from unittest import TestCase
from unittest.mock import MagicMock, PropertyMock, patch


class TestBase(TestCase):
    """Base class for unit tests."""

    # Paths in this list will be automatically patched for all test cases.
    # Overwrite in subclasses to populate the list.
    _to_patch: List[str] = []

    def __init__(self, *args: str, **kwargs: str):
        """Initialize a TestBase instance.

        Args
            args: positional arguments for unittest.TestCase
            kwargs: keyword arguments for unittest.TestCase

        """
        super().__init__(*args, **kwargs)

        self._mocks: Dict[str, MagicMock]

    def setUp(self):
        self._mocks = {}
        for path in self._to_patch:
            if path.endswith('#PROPERTY'):
                path, _ = path.split('#PROPERTY')
                name = path.split('.')[-1]
                patcher = patch(path, new_callable=PropertyMock)
                prop_mock = patcher.start()
                self._mocks[name] = prop_mock
            else:
                patcher = patch(path)
                name = path.split('.')[-1]
                self._mocks[name] = patcher.start()
            self.addCleanup(patcher.stop)



#!/usr/bin/env python3
"""Compile pinned dependencies with pip-compile."""
import subprocess
from pathlib import Path


_DEPS_ORDER = [
    'test-requirements.in', 'dev-requirements.in'
]


this_file = Path(__file__)
repo_root = this_file.parents[1]
deps_dir = repo_root / 'requirements'

# The --allow-unsafe flag lets us pin setuptools.
# The flag will be soon deprecated and made the default behaviour.
# More info: https://github.com/jazzband/pip-tools/issues/989
# TODO (abiro) remove --allow-unsafe once it's deprecated
cmd_template = 'pip-compile -q --allow-unsafe --generate-hashes ' \
               '--output-file {out_file} {in_file}'

for dep in _DEPS_ORDER:
    p = deps_dir / dep
    out_file = p.with_suffix('.txt')
    cmd = cmd_template.format(in_file=p, out_file=out_file)
    subprocess.run(cmd.split(' '), check=True)



#!/usr/bin/env python3
"""Create DynamoDB table for integration tests.

Run from repo root.
"""
import argparse
import logging
from pathlib import Path

import boto3


def create_stack(stack_name: str, template_path: Path):
    logging.info(f'Using template {template_path}')
    with open(template_path) as f:
        template_body = f.read()
    logging.info(f'Creating stack "{stack_name}"...')
    client = boto3.client('cloudformation')
    client.create_stack(
        StackName=stack_name,
        TemplateBody=template_body,
        Capabilities=['CAPABILITY_IAM']
    )
    waiter = client.get_waiter('stack_create_complete')
    waiter.wait(
        StackName=stack_name,
        WaiterConfig={
            'Delay': 10  # seconds
        }
    )
    logging.info(f'Successfully created stack {stack_name}')


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('-n', '--stack_name', type=str,
                        default='DokklibDB-IntegrationTest',
                        help='Stack name')
    parser.add_argument('-p', '--template_path', type=Path,
                        default='./tests/integration/cloudformation.yml',
                        help='Path to the template file name')
    args = parser.parse_args()

    create_stack(args.stack_name, args.template_path)



"""Types for entity names and keys in the single table.

Each item in the table is uniquely identified by its primary key.
The primary key is a composite of the partition key and the sort key.
The partition key for an item is the concatenation of its entity type and the
key value: `ENTITY#key-value`.
The sort key equals the partition key if the item does not model a relation and
`OTHER_ENTITY#other_key` if the item models a relation.
Sort keys without values may be used to query relations with a certain
entity type. For example, to query all the subscriptions of a user, one
would use the `USER#foo@example.com` partition key and the `SUBSCRIPTION#`
sort key.

"""
from abc import ABC
from typing import Any, Dict, List, Optional, Tuple, Type, Union, cast

from dokklib_db_extended.index import GlobalIndex
from dokklib_db_extended.serializer import Serializer

AnySortKey = Union['SortKey', 'PrefixSortKey']


class EntityName(ABC):
    """Abstract base class of entity names.

    Applications must define their entities by inheriting from this class.
    Eg. in "app/entities.py":

    ```python
    import dokklib_db_extended as db

    class User(db.EntityName):
        pass

    class Product(db.EntityName):
        pass

    ...

    ```

    """

    def __new__(cls) -> 'EntityName':  # pragma: no cover
        """Prevent creating abstract base class."""
        raise TypeError(f'{cls.__name__} can not be instantiated.')

    @classmethod
    def to_prefix(cls) -> str:
        """Convert class name to key prefix.

        Returns:
            The key prefix. Eg. if class name is 'User', then the prefix is
            'USER#'.

        """
        if cls is EntityName:
            raise TypeError(f'Entity names must inherit from {cls.__name__}.')  # pragma: no cover  # noqa 501
        return cls.__name__.upper() + '#'


class EntityKey(ABC):
    """Abstract base class of table keys."""

    def __init__(self, entity_name: Type[EntityName], value: str):
        """Initialize an EntityKey instance.

        Args:
            entity_name: The entity type name.
            value: The key value.

        """
        self._prefix = entity_name.to_prefix()
        self._value = value

    # New must match init + subclasses' init as well.
    def __new__(cls, *args: List[Any], **kwargs: Dict[str, Any]) \
            -> 'EntityKey':
        """Prevent creating abstract base class."""
        if cls is EntityKey:
            raise TypeError(f'{EntityKey.__name__} can not be instantiated.')  # pragma: no cover  # noqa 501
        return cast(EntityKey, object.__new__(cls))

    def __str__(self) -> str:
        """Get the string representation."""
        # Eg. ENTITY#value
        return f'{self._prefix}{self._value}'

    def __hash__(self) -> int:
        """Get the hash value."""
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        """Compare semantic equality."""
        return str(self) == str(other)

    @property
    def prefix(self) -> str:
        """Get the entity prefix of the key."""
        return self._prefix

    @property
    def value(self) -> Optional[str]:
        """Get the value of the key."""
        return self._value


class PartitionKey(EntityKey):
    """Partition key."""


class SortKey(EntityKey):
    """Sort key with a value."""


# Shouldn't inherit from `SortKey` as `PrefixSortKey` shouldn't pass where a
# `SortKey` is required.
class PrefixSortKey(EntityKey):
    """Prefix only sort key to query relations."""

    def __init__(self, entity_name: Type[EntityName], value: str = ''):
        """Initialize a PrefixSortKey instance.

        Args:
            entity_name: The entity type name.
            value: Optional prefix value.

        """
        super().__init__(entity_name, value)


class PrimaryKey:
    """Primary (composite) key of a DynamoDB item."""

    def __init__(self, partition_key: PartitionKey, sort_key: SortKey):
        """Initialize a PrimaryKey instance."""
        super().__init__()

        self._pk = partition_key
        self._sk = sort_key
        self._serializer = Serializer()

    def __hash__(self) -> int:
        return hash(self._tuple)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._tuple == other._tuple
        else:
            return self._tuple == other

    @property
    def _tuple(self) -> Tuple[str, str]:
        return str(self.partition_key), str(self.sort_key)

    @property
    def partition_key(self) -> PartitionKey:  # pragma: no cover
        """Get the partition key."""
        return self._pk

    @property
    def sort_key(self) -> SortKey:  # pragma: no cover
        """Get the sort key."""
        return self._sk

    def serialize(self, global_index: GlobalIndex) -> Dict[str, Any]:
        """Serialize the primary key to a DynamoDB item.

        Args:
            global_index: The global index where this key will be used.

        Returns:
            The serialized key.

        """
        pk_name = global_index.partition_key
        sk_name = global_index.sort_key
        item = {
            pk_name: str(self.partition_key),
            sk_name: str(self.sort_key)
        }
        return self._serializer.serialize_dict(item)



# flake8: noqa
# mypy: implicit-reexport

# Flake 8 would complain about unused imports if it was enabled on this file.

from dokklib_db_extended.errors.exceptions import *
from dokklib_db_extended.errors.client import ClientError
from dokklib_db_extended.errors.transaction import TransactionCanceledException



import dokklib_db_extended as db
import dokklib_db_extended.keys as m

from tests.unit import TestBase


class Project(db.EntityName):
    pass


class Session(db.EntityName):
    pass


class Subscription(db.EntityName):
    pass


class User(db.EntityName):
    pass


class TestToPrefix(TestBase):
    def test_correct_prefix(self):
        self.assertEqual(User.to_prefix(), 'USER#')

    def test_pk_prefix(self):
        pk = m.PartitionKey(User, 'val')
        self.assertEqual(User.to_prefix(), pk.prefix)


class TestEntityKeyEq(TestBase):
    def test_self(self):
        pk = m.PartitionKey(User, 'value')
        self.assertEqual(pk, pk)

    def test_eq(self):
        pk_1 = m.PartitionKey(User, 'value')
        pk_2 = m.PartitionKey(User, 'value')
        self.assertEqual(pk_1, pk_2)

    def test_pk_eq_sq(self):
        pk = m.PartitionKey(User, 'value')
        sk = m.SortKey(User, 'value')
        self.assertEqual(pk, sk)


class TestEntityKeyHash(TestBase):
    def test_hash_eq(self):
        pk_1 = m.PartitionKey(User, 'value')
        pk_2 = m.PartitionKey(User, 'value')
        self.assertEqual(hash(pk_1), hash(pk_2))


class TestStrValueMixin:
    _constructor = None

    def test_value(self):
        value = 'value'
        pk = self._constructor(Project, value)
        self.assertEqual(pk, f'PROJECT#{value}')

    def test_different_types(self):
        value = 'value'
        pk_domain = self._constructor(Project, value)
        pk_user = self._constructor(User, value)
        self.assertNotEqual(pk_domain, pk_user)


class TestPartitionKeyStr(TestBase, TestStrValueMixin):
    _constructor = m.PartitionKey


class TestSortKeyStr(TestBase, TestStrValueMixin):
    _constructor = m.SortKey


class TestPrefixSortKeyStr(TestBase):
    def test_no_value(self):
        sk_domain = m.PrefixSortKey(Subscription)
        self.assertEqual(sk_domain, 'SUBSCRIPTION#')


class TestEntityKeyRepr(TestBase):
    def test_pk_repr_no_leak(self):
        """Representation of PK shouldn't leak DB data."""
        value = 'pk-value-1234'
        pk = m.PartitionKey(Session, value)
        self.assertNotIn(value, repr(pk))

    def test_sk_repr_no_leak(self):
        """Representation of SK shouldn't leak DB data."""
        value = 'sk-value-5678'
        sk = m.SortKey(Session, value)
        self.assertNotIn(value, repr(sk))


class TestPrimaryKey(TestBase):
    def setUp(self):
        self._pk = m.PartitionKey(User, 'alice')
        self._sk = m.SortKey(Project, 'foo')
        self._primary = m.PrimaryKey(self._pk, self._sk)

    def test_repr_no_leak(self):
        """Representation of primary key shouldn't leak DB data."""
        self.assertNotIn(self._pk.value, repr(self._primary))
        self.assertNotIn(self._sk.value, repr(self._primary))

    def test_self_eq(self):
        p = m.PrimaryKey(self._pk, self._sk)
        return self.assertEqual(self._primary, p)

    def test_str_eq(self):
        t = (str(self._pk), str(self._sk))
        return self.assertEqual(self._primary, t)

    def test_self_hash(self):
        p = m.PrimaryKey(self._pk, self._sk)
        return self.assertEqual(hash(self._primary), hash(p))

    def test_str_hash(self):
        t = (str(self._pk), str(self._sk))
        return self.assertEqual(hash(self._primary), hash(t))

    def test_serialize(self):
        index = db.PrimaryGlobalIndex()
        res = self._primary.serialize(index)
        des_res = self._primary._serializer.deserialize_dict(res)
        exp = {
            index.partition_key: str(self._pk),
            index.sort_key: str(self._sk)
        }
        self.assertDictEqual(des_res, exp)



#!/usr/bin/env python3
"""Generate DynamoDB exceptions.

Boto3 creates exceptions dynamically from a service model which is not
desirable, because it breaks static analysis. This script generates a Python
file from the dynamically loaded Boto3 exceptions and puts it in
dokklib_db_extended/exceptions.py

"""
import boto3


IGNORE_EXCEPTIONS = {
    # We provide our own ClientError implementation for type checking coverage
    'ClientError',
    # We provide our own TransactionCanceledException that contains
    # cancellation reasons.
    'TransactionCanceledException'
}

MISSING_EXCEPTIONS = {
    # These exceptions are not raised by Boto3, but they are returned as
    # transaction cancellation reasons.
    'ThrottlingError',
    'ValidationError'
}


def _get_exception(name):
    return [
        f'class {name}(ClientError):',
        '    """Please check DynamoDB docs for documentation."""',
        '',
        ''
    ]


# Must match earliest support version
assert boto3.__version__ == '1.10.34', boto3.__version__


lines = [
    '"""Autogenerated DynamoDB exceptions.',
    '',
    'This file was autogenerated by scripts/generate_exceptions.py.',
    'Do not edit it manually!',
    '',
    '"""',
    'from dokklib_db_extended.errors.client import ClientError',
    '',
    ''
]

client = boto3.client('dynamodb')
for name in dir(client.exceptions):
    if name[0].isupper() and name not in IGNORE_EXCEPTIONS:
        assert name not in MISSING_EXCEPTIONS, name
        lines.extend(_get_exception(name))

for name in MISSING_EXCEPTIONS:
    lines.extend(_get_exception(name))

# Remove duplicate empty line
lines = lines[:-1]
with open('dokklib_db_extended/errors/exceptions.py', 'w') as f:
    f.write('\n'.join(lines))



import dokklib_db_extended as db
from dokklib_db_extended.errors.transaction import TransactionCanceledException

from tests.unit import TestBase


class TestTransactionCanceledException(TestBase):
    _op_name = 'TransactWriteItems'

    def _get_error(self, msg):
        return {
            'Error': {
                'Message': msg,
                'Code': 'TransactionCanceledException'
            }
        }

    def test_empty_message(self):
        error = self._get_error('')
        e = TransactionCanceledException([], '', error, self._op_name)
        self.assertListEqual(e.reasons, [])

    def test_mismatch(self):
        error = self._get_error('')
        e = TransactionCanceledException(['1'], '', error, self._op_name)
        with self.assertRaises(ValueError):
            e.reasons

    def test_one_reason(self):
        msg = 'Transaction cancelled, please refer cancellation reasons for ' \
              'specific reasons [ConditionalCheckFailed]'
        error = self._get_error(msg)
        e = TransactionCanceledException(['1'], '', error, self._op_name)
        exp = [db.errors.ConditionalCheckFailedException]
        self.assertListEqual(e.reasons, exp)

    def test_two_reasons(self):
        msg = 'Transaction cancelled, please refer cancellation reasons for ' \
              'specific reasons [ConditionalCheckFailed, None]'
        error = self._get_error(msg)
        e = TransactionCanceledException(['oparg1', 'oparg2'],
                                         '',
                                         error,
                                         self._op_name)
        exp = [db.errors.ConditionalCheckFailedException, None]
        self.assertListEqual(e.reasons, exp)

    def test_no_space_reasons(self):
        msg = 'Transaction cancelled, please refer cancellation reasons for ' \
              'specific reasons [ConditionalCheckFailed,None]'
        error = self._get_error(msg)
        e = TransactionCanceledException(['oparg1', 'oparg2'],
                                         '',
                                         error,
                                         self._op_name)
        exp = [db.errors.ConditionalCheckFailedException, None]
        self.assertListEqual(e.reasons, exp)

    def test_has_error(self):
        msg = 'Transaction cancelled, please refer cancellation reasons for ' \
              'specific reasons [ConditionalCheckFailed, None]'
        error = self._get_error(msg)
        e = TransactionCanceledException(['oparg1', 'oparg2'],
                                         '',
                                         error,
                                         self._op_name)
        self.assertTrue(e.has_error(db.errors.ConditionalCheckFailedException))

    def test_has_no_error(self):
        msg = 'Transaction cancelled, please refer cancellation reasons for ' \
              'specific reasons [ConditionalCheckFailed, None]'
        error = self._get_error(msg)
        e = TransactionCanceledException(['oparg1', 'oparg2'],
                                         '',
                                         error,
                                         self._op_name)
        self.assertFalse(e.has_error(db.errors.ValidationError))



import re
from typing import Any, Dict, List, Optional, Type

from dokklib_db_extended.errors import exceptions as ex
from dokklib_db_extended.errors.client import ClientError
from dokklib_db_extended.op_args import OpArg


CancellationReasons = List[Optional[Type[ClientError]]]


class TransactionCanceledException(ClientError):
    """The entire transaction request was canceled.

    Please see DynamoDB docs for details.
    https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_TransactWriteItems.html

    """

    # Example match: "reasons [ConditionalCheckFailed, None]"
    _reasons_re = re.compile(r'reasons\W+\[([A-Za-z0-9, ]+)]', re.MULTILINE)

    _codes_to_exceptions: Dict[str, Type[ClientError]] = {
        'ConditionalCheckFailed': ex.ConditionalCheckFailedException,
        'ItemCollectionSizeLimitExceeded': ex.ItemCollectionSizeLimitExceededException,  # noqa 501
        'TransactionConflict': ex.TransactionConflictException,
        'ProvisionedThroughputExceeded': ex.ProvisionedThroughputExceededException,  # noqa 501
        'ThrottlingError': ex.ThrottlingError,
        'ValidationError': ex.ValidationError
    }

    def __init__(self, op_args: List[OpArg], *args: Any, **kwargs: Any):
        """Initialize a TransactionCanceledException instance.

        Args:
            op_args: The list of operations that were the inputs to this
                transaction.

        """
        super().__init__(*args, **kwargs)
        self._op_args = list(op_args)
        self._reasons: Optional[CancellationReasons] = None

    def _extract_reasons(self, message: str) -> List[str]:
        match = re.search(self._reasons_re, message)
        if not match:
            return []
        else:
            reasons = match.group(1)
            split = reasons.split(', ')
            if split[0] == reasons:
                return reasons.split(',')
            else:
                return split

    def _get_reasons(self) -> CancellationReasons:
        db_error = self.response.get('Error', {})
        message = db_error.get('Message', '')
        reasons = self._extract_reasons(message)
        res: CancellationReasons = []
        for r in reasons:
            if r == 'None':
                res.append(None)
            else:
                exception = self._codes_to_exceptions.get(r, ClientError)
                res.append(exception)
        if len(res) != len(self.op_args):
            msg = f'Transaction cancellation reasons don\'t match ' \
                  f'transaction arguments in error:\n{message}'
            raise ValueError(msg)
        return res

    @property
    def op_args(self) -> List[OpArg]:
        """Get the list of inputs to the transaction."""
        return self._op_args

    @property
    def reasons(self) -> CancellationReasons:
        """List of cancellation reasons for each item in the transaction.

        Corresponds to order of `op_args`.

        """
        if self._reasons is None:
            self._reasons = self._get_reasons()
        return self._reasons

    def has_error(self, exception: Type[ClientError]) -> bool:
        """Whether the transaction failed due to a particular exception.

        Args:
            exception: The exception type to check for, eg. `ValidationError`.

        Returns:
            True if any of the failure reasons match the exception type.

        """
        for r in self.reasons:
            if r is exception:
                return True
        else:
            return False



from setuptools import find_packages, setup

with open('README.md') as f:
    long_description = f.read()

def local_scheme(version):
    return ""
setup(
    name='dokklib_db_extended',
    version='0.1.3',
    author='Agost Biro, Cristian Dominguez',
    author_email='agost+dokklib_db_extended@dokknet.com, crisdomgo@gmail.com',
    description='DynamoDB Single Table Library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cdominguezg/dokklib-db-extended',
    download_url='https://github.com/cdominguezg/dokklib-db-extended/archive/v0.1.0.tar.gz',
    packages=find_packages(exclude=['tests*']),
    use_scm_version=False,
    # Needed to let mypy use package for type hints
    zip_safe=False,
    package_data={"dokklib_db_extended": ["py.typed"]},
    setup_requires=['setuptools_scm'],
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=[
        'typing-extensions>=3.7.2,<4',
        'boto3'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Database'
    ]
)



import re
from unittest.mock import patch

import boto3.dynamodb.conditions as cond

import dokklib_db_extended as db
import dokklib_db_extended.op_args as m

from tests.unit import TestBase


class User(db.EntityName):
    pass


class Subscription(db.EntityName):
    pass


class TestOpArg(TestBase):
    def test_iso_now(self):
        res = m.OpArg._iso_now()
        iso_format = r'\d{4}-\d{2}-\d{2}T\d{2}\:\d{2}\:\d{2}'
        self.assertTrue(re.match(iso_format, res))


class OpTestMixin:
    def _get_kwargs(self):
        return self._op_arg.get_kwargs(self._table_name, self._primary_index)

    def setUp(self):
        self._pk = db.PartitionKey(User, 'eva.lu-ator@example.com')
        self._sk = db.SortKey(Subscription, 'mitpress.mit.edu')
        self._table_name = 'my-table'
        self._primary_index = db.PrimaryGlobalIndex()

    def test_table_name(self):
        kwargs = self._get_kwargs()
        self.assertEqual(kwargs['TableName'], self._table_name)


class ConsistencyTestMixin:
    def test_key(self):
        kwargs = self._get_kwargs()
        self.assertFalse(kwargs['ConsistentRead'])


class KeyTestMixin:
    def test_key(self):
        kwargs = self._get_kwargs()
        key = kwargs['Key']
        self.assertEqual(key['PK']['S'], str(self._pk))
        self.assertEqual(key['SK']['S'], str(self._sk))


class TestDeleteArg(KeyTestMixin, OpTestMixin, TestBase):
    def setUp(self):
        super().setUp()
        self._op_arg = m.DeleteArg(self._pk, self._sk)

    def test_not_idempotent(self):
        op_arg = m.DeleteArg(self._pk, self._sk, idempotent=False)
        kwargs = op_arg.get_kwargs(self._table_name, self._primary_index)
        self.assertEqual(kwargs['ConditionExpression'],
                         'attribute_exists(PK)')

    def test_idempotent(self):
        kwargs = self._get_kwargs()
        self.assertNotIn('ConditionExpression', kwargs)


class TestGetArg(ConsistencyTestMixin, KeyTestMixin, OpTestMixin, TestBase):
    def setUp(self):
        super().setUp()
        self._op_arg = m.GetArg(self._pk, self._sk)

    def test_projection(self):
        op_arg = m.GetArg(self._pk, self._sk, attributes=['PK', 'SK', 'foo'])
        kwargs = op_arg.get_kwargs(self._table_name, self._primary_index)
        proj = kwargs['ProjectionExpression']
        self.assertLessEqual(proj, 'PK,SK,foo')


class TestPutArg(OpTestMixin, TestBase):
    def setUp(self):
        super().setUp()
        self._op_arg = m.PutArg(self._pk, self._sk)

    @patch('dokklib_db_extended.op_args.PutArg._iso_now')
    def test_adds_created_at(self, iso_now):
        exp_created_at = 'test-time-stamp'
        iso_now.return_value = exp_created_at
        item = self._op_arg._get_dynamo_item(self._primary_index)
        self.assertEqual(item['CreatedAt']['S'], exp_created_at)

    def test_keys_added(self):
        item = self._op_arg._get_dynamo_item(self._primary_index)
        self.assertEqual(item['PK']['S'], self._pk)
        self.assertEqual(item['SK']['S'], self._sk)

    def test_adds_attributes(self):
        put_arg = m.PutArg(self._pk, self._sk,
                           attributes={'foo': '1', 'bar': 2})
        item = put_arg._get_dynamo_item(self._primary_index)
        self.assertEqual(item['foo']['S'], '1')
        self.assertEqual(item['bar']['N'], '2')

    def test_attributes_dont_overwrite_keys(self):
        attributes = {
            'foo': '1',
            'bar': 2,
            'PK': 'my-pk',
            'SK': 'my-sk'
        }
        put_arg = m.PutArg(self._pk, self._sk, attributes=attributes)
        item = put_arg._get_dynamo_item(self._primary_index)
        self.assertEqual(item['PK']['S'], self._pk)
        self.assertEqual(item['SK']['S'], self._sk)

    def test_disallow_overwrite(self):
        put_arg = m.PutArg(self._pk, self._sk, allow_overwrite=False)
        kwargs = put_arg.get_kwargs(self._table_name, self._primary_index)
        cond_expression = 'attribute_not_exists(PK)'
        self.assertEqual(kwargs['ConditionExpression'], cond_expression)


class TestInsertArg(OpTestMixin, TestBase):
    def setUp(self):
        super().setUp()
        self._op_arg = m.InsertArg(self._pk, self._sk)

    def test_no_overwrite(self):
        kwargs = self._get_kwargs()
        cond_expression = 'attribute_not_exists(PK)'
        self.assertEqual(kwargs['ConditionExpression'], cond_expression)


class TestQueryArg(ConsistencyTestMixin,
                   OpTestMixin,
                   TestBase):
    def setUp(self):
        super().setUp()
        self._cond = cond.Key('PK').eq(str(self._pk))
        self._op_arg = m.QueryArg(self._cond)

    def test_key_cond(self):
        kwargs = self._get_kwargs()
        key_cond = kwargs['KeyConditionExpression']
        tokens = key_cond.split(' ')
        self.assertEqual(tokens[1], '=')

    def test_expr_attribute_names(self):
        kwargs = self._get_kwargs()
        key_cond = kwargs['KeyConditionExpression']
        tokens = key_cond.split(' ')
        attr_names = kwargs['ExpressionAttributeNames']
        key, val = list(attr_names.items())[0]
        self.assertEqual(tokens[0], key)
        self.assertEqual(val, 'PK')

    def test_expr_attribute_values(self):
        kwargs = self._get_kwargs()
        key_cond = kwargs['KeyConditionExpression']
        tokens = key_cond.split(' ')
        attr_vals = kwargs['ExpressionAttributeValues']
        key, val = list(attr_vals.items())[0]
        self.assertEqual(tokens[2], key)
        self.assertEqual(val['S'], str(self._pk))

    def test_limit(self):
        limit = 10
        op_arg = m.QueryArg(self._cond, limit=limit)
        kwargs = op_arg.get_kwargs(self._table_name, self._primary_index)
        self.assertLessEqual(kwargs['Limit'], limit)

    def test_default_limit(self):
        kwargs = self._get_kwargs()
        limit = kwargs['Limit']
        self.assertLessEqual(limit, 1000)

    def test_over_limit(self):
        with self.assertRaises(ValueError):
            m.QueryArg(self._cond, limit=10000)

    def test_default_projection(self):
        kwargs = self._get_kwargs()
        proj = kwargs['ProjectionExpression']
        self.assertLessEqual(proj, 'SK')

    def test_projection(self):
        op_arg = m.QueryArg(self._cond, attributes=['PK', 'SK', 'foo'])
        kwargs = op_arg.get_kwargs(self._table_name, self._primary_index)
        proj = kwargs['ProjectionExpression']
        self.assertLessEqual(proj, 'PK,SK,foo')


class TestUpdateArg(OpTestMixin, TestBase):
    def setUp(self):
        super().setUp()
        self._op_arg = m.UpdateArg(self._pk, self._sk)

    def test_key(self):
        kwargs = self._get_kwargs()
        key = kwargs['Key']
        self.assertEqual(key['PK']['S'], str(self._pk))
        self.assertEqual(key['SK']['S'], str(self._sk))

    def test_put_args(self):
        put_attrs = {'foo': 1}
        op_arg = m.UpdateArg(self._pk, self._sk, attr_updates=put_attrs)
        kwargs = op_arg.get_kwargs(self._table_name, self._primary_index)
        foo_update = kwargs['AttributeUpdates']['foo']
        self.assertEqual(foo_update['Action'], 'PUT')
        self.assertEqual(foo_update['Value']['N'], str(put_attrs['foo']))
